import sympy as sp
from rigidbodysimulator.rigid_body_dynamics import S,N
from rigidbodysimulator.substitute_dynamic_symbols import lambdify

import ipyvolume as ipv
import ipywidgets as widgets
import numpy as np
import pandas as pd

p_x = sp.Symbol('p_x')
p_y = sp.Symbol('p_y')
p_z = sp.Symbol('p_z')
P = p_x*S.x + p_y*S.y + p_z*S.z

rotation_lambda = lambdify(P.to_matrix(N))


def rotate2(p, phi_=0, theta_=0, psi_=0):
    args = {
        'p_x': p[0],
        'p_y': p[1],
        'p_z': p[2],
        'phi': phi_,
        'theta': theta_,
        'psi': psi_
    }

    result = rotation_lambda(**args)
    return result.reshape(3, result.shape[2]).T




class TrackPlot3dWidget:
    def __init__(self, df: pd.DataFrame):
        self.figure = ipv.figure(animation=0., animation_exponent=0, width=800, height=800)
        #self.figure.camera_control = 'orbit'

        self.lpp = lpp = 1

        x_min = np.min([-lpp, df['x0'].min()])
        x_max = np.max([lpp, df['x0'].max()])
        y_min = np.min([-lpp, df['y0'].min()])
        y_max = np.max([lpp, df['y0'].max()])
        z_min = np.min([-lpp, df['z0'].min()])
        z_max = np.max([lpp, df['z0'].max()])

        ipv.xlim(x_min, x_max)
        ipv.ylim(y_min, y_max)
        ipv.zlim(z_min, x_max)
        ipv.style.box_off()

        self.track = ipv.pylab.plot(df['x0'].values, df['y0'].values, df['z0'].values)

        self.load_ship_geometry()
        self.port = ipv.plot_mesh(self.X, self.Y_port, self.Z, wireframe=False)
        self.stbd = ipv.plot_mesh(self.X, self.Y_starboard, self.Z, wireframe=False)

        ipv.pylab.view(azimuth=180, elevation=60, distance=1)
        ipv.show()

        self.df = df

        min = df.index[0]
        max = df.index[-1]
        step = (max-min)/100
        self.ui = widgets.interact(self.update,
                                   t=widgets.FloatSlider(value=0, min=min, max=max, step=step))

    def load_ship_geometry(self):
        self.N = 10
        L = self.lpp
        B = 0.1 * self.lpp
        T = 0.05 * self.lpp

        x = np.linspace(0, L, self.N)
        y = np.linspace(0, B, self.N)
        z = np.linspace(0, T, self.N)

        X, Y = np.meshgrid(x, y)
        Z = T - np.array(Y)
        Y *= (L - X) ** (0.4)

        self.X = X - L / 2
        self.Z = Z
        self.Y_port = Y
        self.Y_starboard = -Y

    def update(self, t):
        index = np.argmin(np.abs(self.df.index - t))
        s = self.df.iloc[index]

        phi = s['phi']
        theta = s['theta']
        psi = s['psi']
        x0 = s['x0']
        y0 = s['y0']
        z0 = s['z0']

        print("t:%0.1f [s]" % t)
        print("phi:%0.1f [deg]" % np.rad2deg(phi))
        print("theta:%0.1f [deg]" % np.rad2deg(theta))
        print("psi:%0.1f [deg]" % np.rad2deg(psi))


        angles = [phi, theta, psi]

        X = self.X.flatten()
        Y = self.Y_starboard.flatten()
        Z = self.Z.flatten()
        result = rotate2(p=[X, Y, Z], phi_=angles[0], theta_=angles[1], psi_=angles[2])
        X_ = result[:, 0] + x0
        Y_ = result[:, 1] + y0
        Z_ = result[:, 2] + z0
        self.stbd.x = X_
        self.stbd.y = Y_
        self.stbd.z = Z_

        X = self.X.flatten()
        Y = self.Y_port.flatten()
        Z = self.Z.flatten()
        result = rotate2(p=[X, Y, Z], phi_=angles[0], theta_=angles[1], psi_=angles[2])
        X_ = result[:, 0] + x0
        Y_ = result[:, 1] + y0
        Z_ = result[:, 2] + z0
        self.port.x = X_
        self.port.y = Y_
        self.port.z = Z_


def track_plot(df, ax, l=1, time_step='1S', **kwargs):
    df.plot(x='y0', y='x0', ax=ax, **kwargs)

    df_ = df.copy()
    df_.index = pd.TimedeltaIndex(df_.index, unit='s')
    df_ = df_.resample(time_step).first()

    def plot_body(row):
        x = row['y0']
        y = row['x0']
        psi = row['psi']
        xs = [x - l / 2 * np.sin(psi), x + l / 2 * np.sin(psi)]
        ys = [y - l / 2 * np.cos(psi), y + l / 2 * np.cos(psi)]
        ax.plot(xs, ys, 'k-')

    for index, row in df_.iterrows():
        plot_body(row)

    ax.set_xlabel('y0')
    ax.set_ylabel('x0')
    ax.axis('equal')


import sympy as sp
import sympy.physics.mechanics as me

def substitute_dynamic_symbols(expression):
    dynamic_symbols = me.find_dynamicsymbols(expression)
    derivatives = find_derivatives(dynamic_symbols)

    derivative_list = []
    # First susbtitute the Derivatives starting with the highest order (since higher order could be broken up in lower order)
    subs = []
    for order in reversed(sorted(derivatives.keys())):
        for derivative in list(derivatives[order]):
            name = find_name(derivative)
            symbol = sp.Symbol(name)
            subs.append((derivative, symbol))
            derivative_list.append(derivative)

    new_expression_derivatives = expression.subs(subs)


    none_derivatives = dynamic_symbols - set(derivative_list)

    # ...Then substitute the dynamic symbols
    subs = []
    for dynamic_symbol in list(none_derivatives):
        name = find_name(dynamic_symbol=dynamic_symbol)
        symbol = sp.Symbol(name)
        subs.append((dynamic_symbol, symbol))

    new_expression = new_expression_derivatives.subs(subs)

    return new_expression


def find_name(dynamic_symbol):
    if isinstance(dynamic_symbol, sp.Derivative):
        name = find_derivative_name(dynamic_symbol)
    else:
        name = dynamic_symbol.name
    return name

def find_derivatives(dynamic_symbols:set)->dict:

    derivatives = {}

    for dynamic_symbol in list(dynamic_symbols):
        if isinstance(dynamic_symbol, sp.Derivative):
            order = dynamic_symbol.args[1][1]

            if not order in derivatives:
                derivatives[order] = []

            derivatives[order].append(dynamic_symbol)

    return derivatives


def find_derivative_name(derivative):

    if not isinstance(derivative, sp.Derivative):
        raise ValueError('%s must be an instance of sympy.Derivative' % derivative)

    order = derivative.args[1][1]
    symbol = derivative.expr

    name = '%s%id' % (symbol.name, order)

    return name


def lambdify(expression):
    new_expression = substitute_dynamic_symbols(expression)
    args = new_expression.free_symbols

    # Rearranging to get the parameters in alphabetical order:
    symbol_dict = {symbol.name: symbol for symbol in args}
    symbols = []
    for symbol_name in sorted(symbol_dict.keys()):
        symbols.append(symbol_dict[symbol_name])

    lambda_function = sp.lambdify(symbols, new_expression, modules='numpy')
    return lambda_function


import sympy as sp
import sympy.physics.mechanics as me
import pandas as pd
import numpy as np

from pydy.codegen.ode_function_generators import generate_ode_function
from scipy.integrate import odeint


x0,y0,z0 = me.dynamicsymbols('x0 y0 z0')
x01d,y01d,z01d = me.dynamicsymbols('x01d y01d z01d')
u,v,w = me.dynamicsymbols('u v w')

phi,theta,psi = me.dynamicsymbols('phi theta psi')
phi1d,theta1d,psi1d = me.dynamicsymbols('phi1d theta1d psi1d')

N = me.ReferenceFrame('N')

S = N.orientnew('S', 'Body', [psi,theta,phi],'ZYX')

M = me.Point('M')
O = M.locatenew('P',0)
M.set_vel(N,0)
O.set_vel(S,u*S.x + v*S.y + w*S.z)
O.v1pt_theory(M,N,S)

velocity_matrix = O.vel(N).to_matrix(N)

mass = sp.symbols('m')

I_xx, I_yy, I_zz = sp.symbols('I_xx, I_yy, I_zz')
body_inertia_dyadic = me.inertia(S, ixx=I_xx, iyy=I_yy, izz=I_zz)

body_central_inertia = (body_inertia_dyadic, O)

body = me.RigidBody('Rigid body', masscenter=O, frame = S,
                      mass=mass, inertia=body_central_inertia)

fx, fy, fz, mx, my, mz = sp.symbols('f_x f_y f_z m_x m_y m_z')
force_vector = fx*S.x + fy*S.y + fz*S.z
torque_vector = mx*S.x + my*S.y + mz*S.z
force = (O, force_vector)
torque = (S, torque_vector)


kinematical_differential_equations = [x0.diff() - velocity_matrix[0],
                                      y0.diff() - velocity_matrix[1],
                                      z0.diff() - velocity_matrix[2],
                                      phi.diff() - phi1d,
                                      theta.diff() - theta1d,
                                      psi.diff() - psi1d,
                                     ]

coordinates = [x0, y0, z0, phi, theta, psi]
speeds = [u, v, w, phi1d, theta1d, psi1d]
kane = me.KanesMethod(N, coordinates, speeds, kinematical_differential_equations)

loads = [force,
         torque]

bodies = [body]
fr, frstar = kane.kanes_equations(bodies, loads)

constants = [I_xx, I_yy, I_zz,mass]

specified = [fx, fy, fz, mx, my, mz]  # External force/torque

right_hand_side = generate_ode_function(kane.forcing_full, coordinates,
                                        speeds, constants,
                                        mass_matrix=kane.mass_matrix_full,specifieds=specified)


def simulate(t, force_torque, I_xx, I_yy, I_zz, mass, initial_coordinates=[0, 0, 0, 0, 0, 0],
             initial_speeds=[0, 0, 0, 0, 0, 0])->pd.DataFrame:
    """
    Simulate a rigid body in 6 degrees of freedom under the influence of external torque and forces applied in the body
    reference frame

    :param t: time vector
    :param force_torque: External force and torque: [fx, fy, fx, mx, my, mz] applied in the body
        reference frame

    :param I_xx: roll inertia [kg*m^2]
    :param I_yy: pitch inertia [kg*m^2]
    :param I_zz: yaw inertia [kg*m^2]
    :param mass: mass of body [kg]
    :param initial_coordinates: [x0, y0, z0, phi, theta, psi]
    :param initial_speeds: [u, v, w, phi1d, theta1d, psi1d]  Where u, v w are speeds in ship fixed system
    :return: Pandas data frame with time series.
    """

    start = np.array(initial_coordinates + initial_speeds)

    numerical_specified = force_torque

    numerical_constants = np.array([I_xx, I_yy, I_zz, mass])

    df = pd.DataFrame(index=t)
    y = odeint(right_hand_side, start, t, args=(numerical_specified, numerical_constants))

    for i, symbol in enumerate(coordinates + speeds):
        name = symbol.name
        df[name] = y[:, i]

    return df






import setuptools
import os
import versioneer
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call

with open("README.md", "r") as fh:
    long_description = fh.read()


directory = os.path.dirname(os.path.abspath(__file__))
requirements_path = (os.path.join(directory,'requirements.txt'))
with open(requirements_path,mode = 'r') as file:
    requirements = file.readlines()

extras = {
    'dev': [
        'wheel','versioneer'
    ]
}

test_requirements = ['pytest']

setuptools.setup(
    name="rigidbodysimulator",
    version=versioneer.get_version(),
    author="SSPA",
    author_email="maa@sspa.se",
    description="rigidbodysimulator simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/martinlarsalbert/motion-to-acceleration",
    packages=['rigidbodysimulator'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',

    install_requires=requirements,
    tests_require = test_requirements,
    extras_require=extras,

    package_data={
        'rigidbodysimulator': [],
      },
    scripts=[]

)